import 'dart:async';

import 'package:back_button_interceptor/back_button_interceptor.dart';
import 'package:flutter/services.dart';

import 'package:flutter/material.dart';
import 'package:flutter_webview_plugin/flutter_webview_plugin.dart';

import 'package:myfatoorah_flutter/model/MFError.dart';
import 'package:myfatoorah_flutter/model/directpayment/MFCardInfo.dart';
import 'package:myfatoorah_flutter/model/directpayment/MFDirectPaymentResponse.dart';
import 'package:myfatoorah_flutter/model/directpayment/SDKDirectPaymentResponse.dart';
import 'package:myfatoorah_flutter/model/executepayment/MFExecutePaymentRequest.dart';
import 'package:myfatoorah_flutter/model/executepayment/SDKExecutePaymentResponse.dart';
import 'package:myfatoorah_flutter/model/MyBaseResponse.dart';
import 'package:myfatoorah_flutter/model/initpayment/MFInitiatePaymentRequest.dart';
import 'package:myfatoorah_flutter/model/initpayment/SDKInitiatePaymentResponse.dart';
import 'package:myfatoorah_flutter/model/paymentstatus/MFPaymentStatusRequest.dart';
import 'package:myfatoorah_flutter/model/paymentstatus/SDKPaymentStatusResponse.dart';
import 'package:myfatoorah_flutter/model/sendpayment/MFSendPaymentRequest.dart';
import 'package:myfatoorah_flutter/model/sendpayment/SDKSendPaymentResponse.dart';
import 'package:myfatoorah_flutter/utils/MFResult.dart';
import 'package:myfatoorah_flutter/utils/APPCONST.dart';

import 'package:http/http.dart' as http;
import 'dart:convert';
import 'dart:io';

// Export all MyFatoorah SDK classes

export 'package:myfatoorah_flutter/model/directpayment/MFCardInfo.dart';
export 'package:myfatoorah_flutter/model/directpayment/MFDirectPaymentResponse.dart';
export 'package:myfatoorah_flutter/model/directpayment/SDKDirectPaymentResponse.dart';
export 'package:myfatoorah_flutter/model/executepayment/MFExecutePaymentRequest.dart';
export 'package:myfatoorah_flutter/model/executepayment/SDKExecutePaymentResponse.dart';
export 'package:myfatoorah_flutter/model/initpayment/MFInitiatePaymentRequest.dart';
export 'package:myfatoorah_flutter/model/initpayment/SDKInitiatePaymentResponse.dart';
export 'package:myfatoorah_flutter/model/MFError.dart';
export 'package:myfatoorah_flutter/model/MyBaseResponse.dart';
export 'package:myfatoorah_flutter/model/paymentstatus/MFPaymentStatusRequest.dart';
export 'package:myfatoorah_flutter/model/paymentstatus/SDKPaymentStatusResponse.dart';
export 'package:myfatoorah_flutter/model/sendpayment/MFSendPaymentRequest.dart';
export 'package:myfatoorah_flutter/model/sendpayment/SDKSendPaymentResponse.dart';

export 'package:myfatoorah_flutter/utils/APPCONST.dart';
export 'package:myfatoorah_flutter/utils/MFResult.dart';
export 'package:myfatoorah_flutter/utils/MFAPILanguage.dart';
export 'package:myfatoorah_flutter/utils/MFCurrencyISO.dart';
export 'package:myfatoorah_flutter/utils/MFKeyType.dart';
export 'package:myfatoorah_flutter/utils/MFMobileISO.dart';
export 'package:myfatoorah_flutter/utils/MFNotificationOption.dart';

var MFSDK = new MyFatoorahFlutter();

class MyFatoorahFlutter implements SDKListener {

  static const MethodChannel _channel =
  const MethodChannel('myfatoorah_flutter');

  static Future<String> get platformVersion async {
    final String version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }

  Function func;
  BuildContext myContext;
  String apiLang = "en";

  AppBarSpecs appBarSpecs = AppBarSpecs();

  void setUpAppBar({title, titleColor, backgroundColor, isShowAppBar}) {

    appBarSpecs = AppBarSpecs();

    if(title != null) appBarSpecs.title = title;
    if(titleColor != null) appBarSpecs.titleColor = titleColor;
    if(backgroundColor != null) appBarSpecs.backgroundColor = backgroundColor;
    if(isShowAppBar != null) appBarSpecs.isShowAppBar = isShowAppBar;
  }

  void init(String baseUrl, String token) {
    APPCONST.BASE_URL = baseUrl;
    APPCONST.TOKEN = "bearer " + token;
  }

  // Send Payment
  void sendPayment(String apiLang, MFSendPaymentRequest request, Function func) async{

    this.apiLang = apiLang;

    http.post(APPCONST.SEND_PAYMENT,
        headers: {
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: request.toJson()).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func(MFResult.fail<MFSendPaymentResponse>(new MFError(statusCode, error.message)));
        return;
      }

      var result = SDKSendPaymentResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      func(MFResult.success(result));
    });
  }


  // Initiate Payment
  void initiatePayment(
      MFInitiatePaymentRequest request, String apiLang, Function func) async{

    this.apiLang = apiLang;

    http.post(APPCONST.INITIATE_PAYMENT,
        headers: {
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: request.toJson()).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func(MFResult.fail<MFInitiatePaymentResponse>(new MFError(statusCode, error.message)));
        return;
      }

      var result = SDKInitiatePaymentResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      func(MFResult.success(result));
    });
  }

  // Execute Payment
  void executePayment(BuildContext context,
      MFExecutePaymentRequest request, String apiLang, Function func) async{

    this.apiLang = apiLang;

    if(request.callBackUrl == null || request.callBackUrl.isEmpty)
      request.callBackUrl = APPCONST.CALL_BACK_URL;

    if(request.errorUrl == null || request.errorUrl.isEmpty)
      request.errorUrl = APPCONST.ERROR_URL;

    http.post(APPCONST.EXECUTE_PAYMENT,
        headers: {
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: request.toJson()).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        print("statusCode: " + statusCode.toString());
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func("", MFResult.fail<MFPaymentStatusResponse>(new MFError(statusCode, error.message)));
        return;
      }

      var result = SDKExecutePaymentResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      if (!result.isDirectPayment) {
        this.func = func;
        this.myContext = context;
        showWebView(result.invoiceId.toString(), result.paymentURL);
      }
      else
        func(result.invoiceId, MFResult.fail<MFPaymentStatusResponse>(new MFError(APPCONST.CONFIG_ERROR, APPCONST.incorrectPaymentMethod)));
    });
  }

  // Execute Direct Payment
  void executeDirectPayment(BuildContext context,
      MFExecutePaymentRequest request, MFCardInfo mfCardInfo, String apiLang,
      Function func) async{

    this.apiLang = apiLang;

    if(request.callBackUrl == null || request.callBackUrl.isEmpty)
      request.callBackUrl = APPCONST.CALL_BACK_URL;

    if(request.errorUrl == null || request.errorUrl.isEmpty)
      request.errorUrl = APPCONST.ERROR_URL;

    http.post(APPCONST.EXECUTE_PAYMENT,
        headers: {
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: request.toJson()).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        print("statusCode: " + statusCode.toString());
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func("", MFResult.fail<MFDirectPaymentResponse>(new MFError(statusCode, error.message)));
        return;
      }

      var result = SDKExecutePaymentResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      if (result.isDirectPayment) {
        directPayment(context, apiLang, result.invoiceId, result.paymentURL, mfCardInfo, func);
      }
      else
        func("", MFResult.fail<MFDirectPaymentResponse>(new MFError(APPCONST.CONFIG_ERROR, APPCONST.accountDoNotHaveDirectPayment)));
    });
  }

  // Execute Direct Payment with Recurring
  void executeDirectPaymentWithRecurring(BuildContext context,
      MFExecutePaymentRequest request, MFCardInfo mfCardInfo, int intervalDays,
      String apiLang, Function func) async{

    mfCardInfo.setRecurringIntervalDays(intervalDays);

    executeDirectPayment(context, request, mfCardInfo, apiLang, func);
  }

  // Direct Payment
  void directPayment(BuildContext context,
      String apiLang, int invoiceId, String paymentURL, MFCardInfo request, Function func) async{

    this.apiLang = apiLang;

    http.post(paymentURL,
        headers: {
          HttpHeaders.contentTypeHeader: ContentType.json.toString(),
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: jsonEncode(request.toJson())).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func(invoiceId.toString(), MFResult.fail<MFDirectPaymentResponse>(new MFError(statusCode, error.message)));
        return;
      }

      var result = SDKDirectPaymentResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      if(request.bypass3DS) {
        var request = MFPaymentStatusRequest(paymentId: result.paymentId)
            .toJson();
        var cardInfoResponse = CardInfoResponse(result.token, result.cardInfo);
        paymentStatus(
            apiLang, request, func, cardInfoResponse: cardInfoResponse,
            invoiceId: invoiceId.toString());
      }
      else {
        this.func = func;
        this.myContext = context;
        showWebView(invoiceId.toString(), result.paymentURL);
      }
    });
  }

  // Payment Status
  void paymentStatus(
      String apiLang, Map request, Function func,
      {bool isFinish, CardInfoResponse cardInfoResponse, invoiceId: String}) async{

    this.apiLang = apiLang;

    http.post(APPCONST.PAYMENT_STATUS,
        headers: {
          HttpHeaders.contentTypeHeader: ContentType.json.toString(),
          HttpHeaders.authorizationHeader: APPCONST.TOKEN,
          HttpHeaders.acceptLanguageHeader: apiLang
        }, body: jsonEncode(request)).then((http.Response response) {

      final int statusCode = response.statusCode;

      if (statusCode < 200 || statusCode >= 400 || json == null) {
        var error = MyBaseResponse.fromJson(json.decode(response.body));
        func(invoiceId, MFResult.fail<MFPaymentStatusResponse>(new MFError(statusCode, error.message)));

        if(isFinish != null && isFinish == true)
          Navigator.pop(myContext);
        return;
      }

      var result = SDKPaymentStatusResponse
          .fromJson(json.decode(response.body)).data;

      print(result.toJson());

      if(cardInfoResponse != null)
        func(invoiceId, MFResult.success(MFDirectPaymentResponse(result, cardInfoResponse)));
      else
        func(invoiceId, MFResult.success(result));

      if(isFinish != null && isFinish == true)
        Navigator.pop(myContext);

    });
  }

  void showWebView(String invoiceId, String paymentURL) async{
    Navigator.push(myContext,
      MaterialPageRoute(builder: (context) => MyApp(invoiceId, paymentURL, this,
          appBarSpecs)),
    );
  }

  @override
  void fetchPaymentStatusByAPI(String invoiceId, MFPaymentStatusRequest request) {
    paymentStatus(apiLang, request.toJson(), func, isFinish: true, invoiceId: invoiceId);
  }

  @override
  void onCancelButtonClicked(String invoiceId) {
    func(invoiceId, MFResult.fail<MFPaymentStatusResponse>(
        new MFError(APPCONST.PAYMENT_CANCELED_ERROR, APPCONST.userCancelPayment)));
    Navigator.pop(myContext);
  }

}

class AppBarSpecs {
  String _title = "MyFatoorah Payment";
  Color _titleColor = Color(0xffffffff);
  Color _backgroundColor = Color(0xff0495ca);
  bool _isShowAppBar = true;

  String get title => _title;

  set title(String value) {
    _title = value;
  }

  Color get titleColor => _titleColor;

  set titleColor(Color value) {
    _titleColor = value;
  }

  Color get backgroundColor => _backgroundColor;

  set backgroundColor(Color value) {
    _backgroundColor = value;
  }

  bool get isShowAppBar => _isShowAppBar;

  set isShowAppBar(bool value) {
    _isShowAppBar = value;
  }

}

abstract class SDKListener {
  void fetchPaymentStatusByAPI(String invoiceId, MFPaymentStatusRequest request);
  void onCancelButtonClicked(String invoiceId);
}

class MyApp extends StatefulWidget {

  var invoiceId = "";
  var paymentURL = "";
  SDKListener sdkListener;
  AppBarSpecs appBarSpecs;

  MyApp(String invoiceId, String paymentURL, SDKListener sdkListener,
      AppBarSpecs appBarSpecs) {

    this.invoiceId = invoiceId;
    this.paymentURL = paymentURL;
    this.sdkListener = sdkListener;
    this.appBarSpecs = appBarSpecs;
  }

  @override
  _MyAppState createState() => new _MyAppState(invoiceId, paymentURL,
      sdkListener, appBarSpecs);
}

class _MyAppState extends State<MyApp> {

  var invoiceId = "";
  var paymentURL = "";
  SDKListener sdkListener;
  var isRunningWebView = false;
  AppBarSpecs appBarSpecs;

  _MyAppState(String invoiceId, String paymentURL,
      SDKListener callBackPaymentStatus, AppBarSpecs appBarSpecs) {
    this.invoiceId = invoiceId;
    this.paymentURL = paymentURL;
    this.sdkListener = callBackPaymentStatus;
    this.isRunningWebView = false;
    this.appBarSpecs = appBarSpecs;
  }

//  String url = "";
  double progress = 0;

  // ignore: prefer_collection_literals
  final Set<JavascriptChannel> jsChannels = [
    JavascriptChannel(
        name: 'Print',
        onMessageReceived: (JavascriptMessage message) {
          print(message.message);
        }),
  ].toSet();

  final flutterWebViewPlugin = FlutterWebviewPlugin();
  StreamSubscription<double> _onProgressChanged;
  StreamSubscription<WebViewStateChanged> _onStateChanged;

  @override
  void initState() {
    super.initState();

    BackButtonInterceptor.add(myInterceptor);

    _onStateChanged = flutterWebViewPlugin.onStateChanged.listen((WebViewStateChanged state) {
      String url = state.url;

      print("initState,onStateChanged,url: " + url);

      if(mounted) {
        if ((url.contains(APPCONST.CALL_BACK_URL) ||
            url.contains(APPCONST.ERROR_URL))
            && !isRunningWebView) {

          isRunningWebView = true;

          Uri uri = Uri.dataFromString(url);
          String paymentId = uri.queryParameters["paymentId"];
          var request = MFPaymentStatusRequest(paymentId: paymentId);
          sdkListener.fetchPaymentStatusByAPI(invoiceId, request);
        }
      }
    });

    _onProgressChanged =
        flutterWebViewPlugin.onProgressChanged.listen((double progress) {
          if (mounted) {
            setState(() {
              this.progress = progress;
            });
          }
        });
  }

  @override
  void dispose() {
    super.dispose();

    BackButtonInterceptor.remove(myInterceptor);

    _onProgressChanged.cancel();
    _onStateChanged.cancel();
  }

  bool myInterceptor(bool stopDefaultButtonEvent) {
    sdkListener.onCancelButtonClicked(invoiceId);
    return true;
  }

  @override
  Widget build(BuildContext context) {

    return MaterialApp(
      home: Scaffold(
        appBar: (appBarSpecs.isShowAppBar || Theme.of(context).platform == TargetPlatform.iOS)? AppBar(
            title: Text(appBarSpecs.title, style: TextStyle(
                color: appBarSpecs.titleColor
            )),
            backgroundColor: appBarSpecs.backgroundColor,
            actionsIconTheme: IconThemeData(
              color: appBarSpecs.titleColor,
            ),
            actions: <Widget>[
              IconButton(
                icon: const Icon(Icons.autorenew),
                onPressed: () {
                  flutterWebViewPlugin.reload();
                },
              ),
              IconButton(
                icon: const Icon(Icons.cancel),
                onPressed: () {
                  sdkListener.onCancelButtonClicked(invoiceId);
                },
              ),
            ]
        ) : null,
        body: Container(
            child: Column(children: <Widget>[
//              Container(
//                padding: EdgeInsets.all(20.0),
//                child: Text(
//                    "CURRENT URL\n${(url.length > 50) ? url.substring(0, 50) + "..." : url}"),
//              ),
              Container(
                  padding: EdgeInsets.fromLTRB(0, 1, 0, 1),
                  child: progress < 1.0
                      ? LinearProgressIndicator(value: progress)
                      : Container()),
              Expanded(
                child: Container(
//                  margin: const EdgeInsets.all(10.0),
//                  decoration:
//                  BoxDecoration(border: Border.all(color: Colors.blueAccent)),
                  child: WebviewScaffold(
                    url: paymentURL,
                    javascriptChannels: jsChannels,
                    mediaPlaybackRequiresUserGesture: false,
//                    appBar: AppBar(
//                      title: const Text('Widget WebView'),
//                    ),
                    withZoom: true,
                    withLocalStorage: true,
                    hidden: true,
                    initialChild: Container(
                      color: Colors.white,
                      child: const Center(
                        child: Text('Waiting...'),
                      ),
                    ),
                  ),
                ),
              ),
            ])),
      ),
    );
  }
}

class MyfatoorahSdk {
  static const MethodChannel _channel =
  const MethodChannel('myfatoorah_flutter');

  static Future<String> get platformVersion async {
    final String version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }
}
